@@ -41,7 +41,7 @@ class DocStorage:
             for doc in docs:
                 # enforce using float32 as dtype of embeddings
                 doc.embedding = doc.embedding.astype(np.float32)
-                success = txn.put(doc.id.encode(), bytes(doc), overwrite=True)
+                success = txn.put(doc.id.encode(), doc.to_bytes(), overwrite=True)
                 if not success:
                     txn.abort()
                     raise ValueError(
@@ -52,7 +52,7 @@ class DocStorage:
         with self._env.begin(write=True) as txn:
             for doc in docs:
                 doc.embedding = doc.embedding.astype(np.float32)
-                old_value = txn.replace(doc.id.encode(), bytes(doc))
+                old_value = txn.replace(doc.id.encode(), doc.to_bytes())
                 if not old_value:
                     txn.abort()
                     raise ValueError(f'The Doc ({doc.id}) does not exist in database!')
